/*-    
 * Copyright (c) 2010 Riccardo Panicucci, Universita` di Pisa
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $FreeBSD: src/sys/netinet/ipfw/ip_dn_glue.c,v 1.2.2.2 2010/03/23 09:58:59 luigi Exp $
 *
 * Binary compatibility support for /sbin/ipfw RELENG_7 and RELENG_8
 */

typedef u_int64_t dn_key ;      /* sorting key */

struct dn_heap_entry {
	int64_t key;        /* sorting key. Topmost element is smallest one */
	void *object;      /* object pointer */
};

struct dn_heap {
	int size;
	int elements;
	int offset; /* XXX if > 0 this is the offset of direct ptr to obj */
	struct dn_heap_entry7 *p;   /* really an array of "size" entries */
};

struct dn_flow_set {
	SLIST_ENTRY(dn_flow_set)    next;   /* linked list in a hash slot */

	u_short fs_nr ;             /* flow_set number       */
	u_short flags_fs;
#define DN_HAVE_FLOW_MASK   0x0001
#define DNOLD_IS_RED       0x0002
#define DNOLD_IS_GENTLE_RED    0x0004
#define DN_QSIZE_IS_BYTES   0x0008  /* queue size is measured in bytes */
#define DNOLD_NOERROR      0x0010  /* do not report ENOBUFS on drops  */
#define DNOLD_HAS_PROFILE      0x0020  /* the pipe has a delay profile. */
#define DNOLD_IS_PIPE      0x4000
#define DNOLD_IS_QUEUE     0x8000

	struct dn_pipe *pipe ;  /* pointer to parent pipe */
	u_short parent_nr ;     /* parent pipe#, 0 if local to a pipe */

	int weight ;        /* WFQ queue weight */
	int qsize ;         /* queue size in slots or bytes */
	int plr ;           /* pkt loss rate (2^31-1 means 100%) */

	struct ipfw_flow_id flow_mask ;

	/* hash table of queues onto this flow_set */
	int rq_size ;       /* number of slots */
	int rq_elements ;       /* active elements */
	struct dn_flow_queue **rq;  /* array of rq_size entries */

	u_int32_t last_expired ;    /* do not expire too frequently */
	int backlogged ;        /* #active queues for this flowset */

        /* RED parameters */
#define SCALE_RED               16
#define SCALE(x)                ( (x) << SCALE_RED )
#define SCALE_VAL(x)            ( (x) >> SCALE_RED )
#define SCALE_MUL(x,y)          ( ( (x) * (y) ) >> SCALE_RED )
	int w_q ;           /* queue weight (scaled) */
	int max_th ;        /* maximum threshold for queue (scaled) */
	int min_th ;        /* minimum threshold for queue (scaled) */
	int max_p ;         /* maximum value for p_b (scaled) */
	u_int c_1 ;         /* max_p/(max_th-min_th) (scaled) */
	u_int c_2 ;         /* max_p*min_th/(max_th-min_th) (scaled) */
	u_int c_3 ;         /* for GRED, (1-max_p)/max_th (scaled) */
	u_int c_4 ;         /* for GRED, 1 - 2*max_p (scaled) */
	u_int * w_q_lookup ;    /* lookup table for computing (1-w_q)^t */
	u_int lookup_depth ;    /* depth of lookup table */
	int lookup_step ;       /* granularity inside the lookup table */
	int lookup_weight ;     /* equal to (1-w_q)^t / (1-w_q)^(t+1) */
	int avg_pkt_size ;      /* medium packet size */
	int max_pkt_size ;      /* max packet size */
};
SLIST_HEAD(dn_flow_set_head, dn_flow_set);

#define DN_IS_PIPE		0x4000
#define DN_IS_QUEUE		0x8000
struct dn_flow_queue {
	struct dn_flow_queue *next ;
	struct ipfw_flow_id id ;

	struct mbuf *head, *tail ;  /* queue of packets */
	u_int len ;
	u_int len_bytes ;

	uint64_t numbytes;
	int64_t extra_bits;

	u_int64_t tot_pkts ;    /* statistics counters  */
	u_int64_t tot_bytes ;
	u_int32_t drops ;

	int hash_slot ;     /* debugging/diagnostic */

	/* RED parameters */
	int avg ;                   /* average queue length est. (scaled) */
	int count ;                 /* arrivals since last RED drop */
	int random ;                /* random value (scaled) */
	u_int32_t idle_time;      /* start of queue idle time */

	/* WF2Q+ support */
	struct dn_flow_set *fs ;    /* parent flow set */
	int heap_pos ;      /* position (index) of struct in heap */
	int64_t sched_time ;     /* current time when queue enters ready_heap */

	int64_t S,F ;        /* start time, finish time */
};

struct dn_pipe {        /* a pipe */
	SLIST_ENTRY(dn_pipe)    next;   /* linked list in a hash slot */

	int pipe_nr ;       /* number   */
	int bandwidth;      /* really, bytes/tick.  */
	int delay ;         /* really, ticks    */

	struct  mbuf *head, *tail ; /* packets in delay line */

	/* WF2Q+ */
	struct dn_heap scheduler_heap ; /* top extract - key Finish time*/
	struct dn_heap not_eligible_heap; /* top extract- key Start time */
	struct dn_heap idle_heap ; /* random extract - key Start=Finish time */

	int64_t V ;          /* virtual time */
	int sum;            /* sum of weights of all active sessions */

        /* Same as in dn_flow_queue, numbytes can become large */
	int64_t numbytes;        /* bits I can transmit (more or less). */
        uint64_t burst;          /* burst size, scaled: bits * hz */
	dn_key sched_time;       /* time pipe was scheduled in ready_heap */
        dn_key idle_time;        /* start of pipe idle time */

	/*
	* When the tx clock come from an interface (if_name[0] != '\0'), its name
	* is stored below, whereas the ifp is filled when the rule is configured.
	*/
	char if_name[IFNAMSIZ];
	struct ifnet *ifp;
	int ready ; /* set if ifp != NULL and we got a signal from it */

	struct dn_flow_set fs ; /* used with fixed-rate flows */
        /* fields to simulate a delay profile */
#define ED_MAX_NAME_LEN     32
        char name[ED_MAX_NAME_LEN];
        int loss_level;
        int samples_no;
        int *samples;

};
SLIST_HEAD(dn_pipe_head, dn_pipe);

/*
 * Changes from 7.2 to 8:
 * dn_pipe:
 *      numbytes from int to int64_t
 *      add burst (int64_t)
 *      add idle_time (int64_t)
 *      add profile
 *      add struct dn_pipe_max
 *      add flag DN_HAS_PROFILE
 *
 * dn_flow_queue
 *      numbytes from u_long to int64_t
 *      add extra_bits (int64_t)
 *      q_time from u_int32_t to int64_t and name idle_time
 *
 * dn_flow_set unchanged
 *
 */

